'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var ironStore = _interopDefault(require('iron-store'));
var cookie = _interopDefault(require('cookie'));

// see https://hapi.dev/family/iron/api/?v=6.0.0#options

const timestampSkewSec = 60;
const defaultOptions = {
  ttl: 15 * 24 * 3600,
  cookieOptions: {
    httpOnly: true,
    secure: true,
    sameSite: "lax",
    path: "/"
  }
};
async function getIronSession(req, res, userOptions) {
  var _userOptions$cookieOp, _userOptions$cookieOp2;

  if (!req || !res || !userOptions || !userOptions.cookieName || !userOptions.password) {
    throw new Error(`next-iron-session: Bad usage. Minimum usage is const session = await getIronSession(req, res, { cookieName: "...", password: "...". Check the usage here: https://github.com/vvo/next-iron-session`);
  }

  const passwordAsAnArray = Array.isArray(userOptions.password) ? userOptions.password : [{
    id: 1,
    password: userOptions.password
  }];
  passwordAsAnArray.forEach(({
    password
  }) => {
    if (password.length < 32) {
      throw new Error(`next-iron-session: Bad usage. Password must be at least 32 characters long.`);
    }
  });
  const options = { ...defaultOptions,
    ...userOptions,
    cookieOptions: { ...defaultOptions.cookieOptions,
      ...(userOptions.cookieOptions || {})
    }
  };

  if (options.ttl === 0) {
    // ttl = 0 means no expiration
    // but in reality cookies have to expire (can't have no max-age)
    // 2147483647 is the max value for max-age in cookies
    // see https://stackoverflow.com/a/11685301/147079
    options.ttl = 2147483647;
  }

  options.cookieOptions.maxAge = computeCookieMaxAge((_userOptions$cookieOp = (_userOptions$cookieOp2 = userOptions.cookieOptions) == null ? void 0 : _userOptions$cookieOp2.maxAge) != null ? _userOptions$cookieOp : options.ttl);
  const store = await getOrCreateStore({
    sealed: cookie.parse(req.headers.cookie || "")[options.cookieName],
    password: passwordAsAnArray,
    ttl: options.ttl * 1000
  });
  return {
    set: store.set,
    get: store.get,
    unset: store.unset,

    async save() {
      const seal = await store.seal();
      const cookieValue = cookie.serialize(options.cookieName, seal, options.cookieOptions);

      if (cookieValue.length > 4096) {
        throw new Error(`next-iron-session: Cookie length is too big ${cookieValue.length}, browsers will refuse it. Try to remove some data.`);
      }

      addToCookies(cookieValue, res);
      return cookieValue;
    },

    destroy() {
      store.clear();
      const cookieValue = cookie.serialize(options.cookieName, "", { ...options.cookieOptions,
        maxAge: 0
      });
      addToCookies(cookieValue, res);
    }

  };
}

function addToCookies(cookieValue, res) {
  var _res$getHeader;

  let existingSetCookie = (_res$getHeader = res.getHeader("set-cookie")) != null ? _res$getHeader : [];

  if (typeof existingSetCookie === "string") {
    existingSetCookie = [existingSetCookie];
  }

  res.setHeader("set-cookie", [...existingSetCookie, cookieValue]);
}

function computeCookieMaxAge(ttl) {
  // The next line makes sure browser will expire cookies before seals are considered expired by the server.
  // It also allows for clock difference of 60 seconds maximum between server and clients.
  // It also makes sure to expire the cookie immediately when value is 0
  return ttl - timestampSkewSec;
}

async function getOrCreateStore({
  sealed,
  password,
  ttl
}) {
  try {
    return await ironStore({
      sealed,
      password,
      ttl
    });
  } catch (error) {
    if (error instanceof Error) {
      if (error.message === "Expired seal" || error.message === "Bad hmac value" || error.message === "Cannot find password: ") {
        // if seal expired or
        // if seal is not valid (encrypted using a different password, when passwords are updated) or
        // if we can't find back the password in the seal
        // then we just start a new session over
        return await ironStore({
          password,
          ttl
        });
      }
    }

    throw error;
  }
}

exports.getIronSession = getIronSession;
//# sourceMappingURL=next-iron-session.cjs.development.js.map
