import { advanceBy, clear } from "jest-date-mock";
import ironStore from "iron-store";

import type { IncomingMessage, ServerResponse } from "http";
import { getIronSession } from "./";

const password = "Gbm49ATjnqnkCCCdhV4uDBhbfnPqsCW0";
const cookieName = "test";

test("no req", async () => {
  // @ts-ignore we actually want to test this
  await expect(getIronSession()).rejects.toThrowErrorMatchingInlineSnapshot(
    `"next-iron-session: Bad usage. Minimum usage is const session = await getSession(req, res, { cookieName: \\"...\\", password: \\"...\\". Check the usage here: https://github.com/vvo/next-iron-session"`,
  );
});

test("no res", async () => {
  // @ts-ignore we actually want to test this
  await expect(getIronSession({})).rejects.toThrowErrorMatchingInlineSnapshot(
    `"next-iron-session: Bad usage. Minimum usage is const session = await getSession(req, res, { cookieName: \\"...\\", password: \\"...\\". Check the usage here: https://github.com/vvo/next-iron-session"`,
  );
});

test("no password", async () => {
  // @ts-ignore we actually want to test this
  await expect(
    getIronSession({}, {}),
  ).rejects.toThrowErrorMatchingInlineSnapshot(
    `"next-iron-session: Bad usage. Minimum usage is const session = await getSession(req, res, { cookieName: \\"...\\", password: \\"...\\". Check the usage here: https://github.com/vvo/next-iron-session"`,
  );
});

test("no cookie name", async () => {
  await expect(
    // @ts-ignore we actually want to test this
    getIronSession({}, {}, { password }),
  ).rejects.toThrowErrorMatchingInlineSnapshot(
    `"next-iron-session: Bad usage. Minimum usage is const session = await getSession(req, res, { cookieName: \\"...\\", password: \\"...\\". Check the usage here: https://github.com/vvo/next-iron-session"`,
  );
});

test("bad password length", async () => {
  await expect(
    // @ts-ignore we actually want to test this
    getIronSession({}, {}, { password: password.substring(1), cookieName }),
  ).rejects.toThrowErrorMatchingInlineSnapshot(
    `"next-iron-session: Bad usage. Password must be at least 32 characters long."`,
  );
});

test("getSession(req, res, {password, cookieName})", async () => {
  await expect(
    getIronSession(
      {
        headers: {},
      } as IncomingMessage,
      {} as ServerResponse,
      { password, cookieName },
    ),
  ).resolves.toMatchInlineSnapshot(`
          Object {
            "destroy": [Function],
            "get": [Function],
            "save": [Function],
            "set": [Function],
            "unset": [Function],
          }
        `);
});

test("session.set", async () => {
  const session = await getIronSession(
    {
      // IncomingMessage always have a headers property
      headers: {},
    } as IncomingMessage,
    {} as ServerResponse,
    { password, cookieName },
  );

  expect(session.set("user", { id: 20 })).toMatchInlineSnapshot(`
    Object {
      "id": 20,
    }
  `);
});

test("session.get", async () => {
  const session = await getIronSession(
    {
      headers: {},
    } as IncomingMessage,
    {} as ServerResponse,
    { password, cookieName },
  );

  session.set("user", { id: 21 });

  expect(session.get("user")).toMatchInlineSnapshot(`
    Object {
      "id": 21,
    }
  `);
});

test("session.unset", async () => {
  const session = await getIronSession(
    {
      headers: {},
    } as IncomingMessage,
    {} as ServerResponse,
    { password, cookieName },
  );

  session.set("user", { id: 22 });
  session.unset("user");
  expect(session.get("user")).toMatchInlineSnapshot(`undefined`);
});

test("session.save", async () => {
  const res = {
    getHeader: jest.fn(),
    setHeader: jest.fn(),
  };

  const session = await getIronSession(
    {
      headers: {},
    } as IncomingMessage,
    res as unknown as ServerResponse,
    { password, cookieName },
  );

  session.set("user", { id: 100 });
  await session.save();
  const headerName = res.setHeader.mock.calls[0][0];
  expect(headerName).toMatchInlineSnapshot(`"set-cookie"`);

  const headerValue = res.setHeader.mock.calls[0][1];
  expect(Array.isArray(headerValue)).toBe(true);
  expect(headerValue).toHaveLength(1);

  const cookie = headerValue[0];
  const seal = cookie.split(";")[0].split("=")[1];
  expect(seal).toHaveLength(284);

  const cookieParams = cookie.split(";").slice(1).join(";");
  expect(cookieParams).toMatchInlineSnapshot(
    `" Max-Age=1295940; Path=/; HttpOnly; Secure; SameSite=Lax"`,
  );
});

test("getSession with existing session cookie", async () => {
  const session = await getIronSession(
    {
      headers: {
        cookie:
          // this seal was previously generated with a ttl of 0
          "test=Fe26.2*1*1e2bacee1edffaeb4a9ba4a07dc36c2c60d20415a60ac1b901033af1f107ead5*LAC9Fn3BJ9ifKMhVL3pP5w*JHhcByIzk4ThLt9rUW-fDMrOwUT7htHy1uyqeOTIqrVwDJ0Bz7TOAwIz_Cos-ug3**7dfa11868bbcc4f7e118342c0280ff49ba4a7cc84c70395bbc3d821a5f460174*6a8FkHxdg322jyym6PwJf3owz7pd6nq5ZIzyLHGVC0c",
      },
    } as IncomingMessage,
    {} as ServerResponse,
    { password, cookieName },
  );

  expect(session.get("user")).toMatchInlineSnapshot(`
    Object {
      "id": 77,
    }
  `);
});

test("When ttl is 0 (infinite session), maxAge is very far away", async () => {
  const res = {
    getHeader: jest.fn(),
    setHeader: jest.fn(),
  };

  const session = await getIronSession(
    {
      headers: {},
    } as IncomingMessage,
    res as unknown as ServerResponse,
    { password, cookieName, ttl: 0 },
  );

  await session.save();
  const headerValue = res.setHeader.mock.calls[0][1];
  const cookie = headerValue[0];
  const maxAgeParam = cookie.split(";")[1].trim();
  expect(maxAgeParam).toMatchInlineSnapshot(`"Max-Age=2147483587"`);
});

test("Passing down a maxAge option", async () => {
  const res = {
    getHeader: jest.fn(),
    setHeader: jest.fn(),
  };

  const session = await getIronSession(
    {
      headers: {},
    } as IncomingMessage,
    res as unknown as ServerResponse,
    {
      password,
      cookieName,
      cookieOptions: {
        maxAge: 1000,
      },
    },
  );

  await session.save();
  const headerValue = res.setHeader.mock.calls[0][1];
  const cookie = headerValue[0];
  const maxAgeParam = cookie.split(";")[1].trim();
  expect(maxAgeParam).toMatchInlineSnapshot(`"Max-Age=940"`);
});

test("session.destroy", async () => {
  const res = {
    getHeader: jest.fn(),
    setHeader: jest.fn(),
  };

  const session = await getIronSession(
    {
      headers: {},
    } as IncomingMessage,
    res as unknown as ServerResponse,
    { password, cookieName },
  );

  session.destroy();
  await session.save();
  expect(session.get()).toMatchInlineSnapshot(`Object {}`);
  expect(res.setHeader.mock.calls[0]).toMatchInlineSnapshot(`
    Array [
      "set-cookie",
      Array [
        "test=; Max-Age=0; Path=/; HttpOnly; Secure; SameSite=Lax",
      ],
    ]
  `);
});

test("When trying to use an expired seal", async () => {
  // 100 seconds seal ttl
  const ttl = 100;

  // iron-store uses miliseconds
  const store = await ironStore({ password, ttl: ttl * 1000 });

  store.set("user", { id: 20 });
  const seal = await store.seal();

  const initialSession = await getIronSession(
    {
      headers: { cookie: `test=${seal}` },
    } as IncomingMessage,
    {} as ServerResponse,
    { password, cookieName, ttl },
  );

  expect(initialSession.get()).toMatchInlineSnapshot(`
    Object {
      "user": Object {
        "id": 20,
      },
    }
  `);

  // advance time by ttl + 60 seconds. THe 60 seconds are coming from the hapijs/iron https://github.com/hapijs/iron/blob/d26fa2e06483a90ee81597815394155529cda798/lib/index.js#L31
  advanceBy((ttl + 60) * 1000);

  const secondSession = await getIronSession(
    {
      headers: { cookie: `test=${seal}` },
    } as IncomingMessage,
    {} as ServerResponse,
    { password, cookieName, ttl },
  );

  // now the session is empty, automatically
  expect(secondSession.get()).toMatchInlineSnapshot(`Object {}`);

  clear();
});

test("When mixing passwords between seals. Example: password was updated server-side without rotation.", async () => {
  // const firstPassword = "Bb0EyombqcDK58k870btymbGJrgZFrN2";
  const secondPassword = "182XhM1mAzottfvzPMN0nh20HMwZprBc";

  const store = await ironStore({ password, ttl: 0 });
  store.set("user", { id: 77 });
  const seal = await store.seal();

  const session = await getIronSession(
    {
      headers: { cookie: `test=${seal}` },
    } as IncomingMessage,
    {} as ServerResponse,
    { password: secondPassword, cookieName },
  );

  expect(session.get("user")).toMatchInlineSnapshot(`undefined`);
});

test("Password rotation", async () => {
  const firstPassword = [
    { id: 1, password: "BcTv8NKLVfGcTt18HqGf2DhEnmJrLbNU" },
  ];
  const secondPassword = [
    // any of these passwords will try to decrypt the seal
    // only the first one will be used to encrypt new seals
    { id: 2, password: "scKVNPWFippYjA3tRjJPuPnK7ocj4Vnn" },
    { id: 1, password: firstPassword[0].password },
  ];

  const store = await ironStore({ password: firstPassword });
  store.set("user", { id: 30 });
  const seal = await store.seal();

  const session = await getIronSession(
    {
      headers: { cookie: `test=${seal}` },
    } as IncomingMessage,
    {} as ServerResponse,
    { password: secondPassword, cookieName },
  );

  expect(session.get("user")).toMatchInlineSnapshot(`
    Object {
      "id": 30,
    }
  `);
});

test("it throws when cookie length is too big", async () => {
  const session = await getIronSession(
    {
      // IncomingMessage always have a headers property
      headers: {},
    } as IncomingMessage,
    {} as ServerResponse,
    { password, cookieName },
  );

  session.set("user", "somevalue".repeat(500));
  await expect(session.save()).rejects.toThrowErrorMatchingInlineSnapshot(
    `"next-iron-session: Cookie length is too big 6341, browsers will refuse it. Try to remove some data."`,
  );
});

test("it keeps previously set cookies (single value)", async () => {
  const res = {
    getHeader: function () {
      return "existingCookie=value";
    },
    setHeader: jest.fn(),
  };

  const session = await getIronSession(
    {
      headers: {},
    } as IncomingMessage,
    res as unknown as ServerResponse,
    { password, cookieName },
  );

  await session.save();

  const headerValue = res.setHeader.mock.calls[0][1];
  expect(headerValue.length).toBe(2);
  expect(headerValue[0]).toBe("existingCookie=value");
});

test("it keeps previously set cookies (multiple values)", async () => {
  const res = {
    getHeader: function () {
      return ["existingCookie=value", "anotherCookie=value2"];
    },
    setHeader: jest.fn(),
  };

  const session = await getIronSession(
    {
      headers: {},
    } as IncomingMessage,
    res as unknown as ServerResponse,
    { password, cookieName },
  );

  await session.save();

  const headerValue = res.setHeader.mock.calls[0][1];
  expect(headerValue.length).toBe(3);
  expect(headerValue[0]).toBe("existingCookie=value");
  expect(headerValue[1]).toBe("anotherCookie=value2");
});

test("it keeps previously set cookies (multiple values) on destroy()", async () => {
  const res = {
    getHeader: function () {
      return ["existingCookie=value", "anotherCookie=value2"];
    },
    setHeader: jest.fn(),
  };

  const session = await getIronSession(
    {
      headers: {},
    } as IncomingMessage,
    res as unknown as ServerResponse,
    { password, cookieName },
  );

  await session.destroy();

  expect(res.setHeader.mock.calls[0]).toMatchInlineSnapshot(`
    Array [
      "set-cookie",
      Array [
        "existingCookie=value",
        "anotherCookie=value2",
        "test=; Max-Age=0; Path=/; HttpOnly; Secure; SameSite=Lax",
      ],
    ]
  `);
});
